@@ -562,6 +562,67 @@ struct socket_info** get_sock_info_list(unsigned short proto)
 	return 0;
 }
 
+/* Check list of active local IPs for grep_sock_info
+ * This function is only used for sockets with the SI_IS_VIRTUAL flag set. This
+ * is so floating (virtual) IPs that are not currently local, are not returned
+ * as matches by grep_sock_info.
+ *
+ * Params:
+ * - si - Socket info of socket that has been flagged with SI_IS_VIRTUAL, 
+ *   that we want to check if it's actually local right now.
+ *
+ * Returns 1 if socket is local, or 0 if not.
+ */
+static int check_local_addresses(struct socket_info* si)
+{
+	int match = 0;
+	struct ifaddrs *ifap, *ifa;
+
+	if (si == NULL) {
+		LM_ERR("Socket info is NULL. Returning no match.\n");
+		return 0;
+	}
+
+	if (!(si->flags & SI_IS_VIRTUAL)) {
+		LM_ERR("Have been passed a socket without the virtual flag set. This should "
+			"not happen. Returning a match to maintain standard behaviour.\n");
+		return 1;
+	}
+
+	if (getifaddrs(&ifap) != 0) {
+		LM_ERR("getifaddrs failed. Assuming no match.\n");
+		return 0;
+	}
+
+	for (ifa = ifap; ifa; ifa = ifa->ifa_next)
+	{
+		/* skip if no IP addr associated with the interface */
+		if (ifa->ifa_addr==0)
+			continue;
+#ifdef AF_PACKET
+		/* skip AF_PACKET addr family since it is of no use later on */
+		if (ifa->ifa_addr->sa_family == AF_PACKET)
+			continue;
+#endif
+		struct ip_addr local_addr;
+		sockaddr2ip_addr(&local_addr, (struct sockaddr*)ifa->ifa_addr);
+
+		LM_DBG("Checking local address: %s\n", ip_addr2a(&local_addr));
+		if (ip_addr_cmp(&si->address, &local_addr)) {
+			match = 1;
+			LM_DBG("Found matching local IP %s for virtual socket %s\n", ip_addr2a(&local_addr), si->name.s);
+			break;
+		}
+	}
+	freeifaddrs(ifap);
+	//Default to not local if no match is found
+	if (!match) {
+		LM_DBG("No matching local IP found for socket %s.\n", si->name.s);
+		return 0;
+	} else {
+		return 1;
+	}
+}
 
 /* helper function for grep_sock_info
  * params:
@@ -653,7 +714,16 @@ retry:
 			}
 			if (si_hname_cmp(&hname, &si->name, &si->address_str,
 								&si->address, si->flags)==0) {
-				goto found;
+				if (si->flags & SI_IS_VIRTUAL) {
+					LM_DBG("Checking virtual socket: [%.*s]\n", si->name.len, si->name.s);
+					if (check_local_addresses(si)) {
+						goto found;
+					} else {
+						LM_DBG("Skipping virtual socket that is not local.\n");
+					}
+				} else {
+					goto found;
+				}
 			}
 			if(si->useinfo.name.s!=NULL) {
 				LM_DBG("checking advertise if host==us:"
@@ -692,6 +762,54 @@ found:
 	return si;
 }
 
+/**
+ *
+ */
+static int _ksr_sockets_no = 0;
+
+/**
+ *
+ */
+int ksr_sockets_no_get(void)
+{
+	return _ksr_sockets_no;
+}
+
+/**
+ *
+ */
+void ksr_sockets_index(void)
+{
+	socket_info_t *si = NULL;
+	struct socket_info** list;
+	unsigned short c_proto;
+
+	if(_ksr_sockets_no > 0) {
+		return;
+	}
+
+	c_proto = PROTO_UDP;
+	do {
+		/* get the proper sock_list */
+		list=get_sock_info_list(c_proto);
+
+		if (list==0) {
+			/* disabled or unknown protocol */
+			continue;
+		}
+
+		for (si=*list; si; si=si->next) {
+			if(si->sockname.s == NULL) {
+				continue;
+			}
+			si->gindex = _ksr_sockets_no;
+			_ksr_sockets_no++;
+		}
+	} while((c_proto = next_proto(c_proto))!=0);
+
+	LM_DBG("number of listen sockets: %d\n", _ksr_sockets_no);
+}
+
 socket_info_t* ksr_get_socket_by_name(str *sockname)
 {
 	socket_info_t *si = NULL;
@@ -726,6 +844,108 @@ socket_info_t* ksr_get_socket_by_name(str *sockname)
 	return NULL;
 }
 
+socket_info_t* ksr_get_socket_by_listen(str *sockstr)
+{
+	socket_info_t *si = NULL;
+	struct socket_info** list;
+	unsigned short c_proto;
+
+	c_proto = PROTO_UDP;
+	do {
+		/* get the proper sock_list */
+		list=get_sock_info_list(c_proto);
+
+		if (list==0) {
+			/* disabled or unknown protocol */
+			continue;
+		}
+
+		for (si=*list; si; si=si->next) {
+			if(si->sock_str.s == NULL) {
+				continue;
+			}
+			LM_DBG("checking if listen %.*s matches %.*s\n",
+					sockstr->len, sockstr->s,
+					si->sock_str.len, si->sock_str.s);
+			if (sockstr->len == si->sock_str.len
+					&& strncasecmp(sockstr->s, si->sock_str.s,
+							sockstr->len)==0) {
+				return si;
+			}
+		}
+	} while((c_proto = next_proto(c_proto))!=0);
+
+	return NULL;
+}
+
+socket_info_t* ksr_get_socket_by_advertise(str *sockstr)
+{
+	socket_info_t *si = NULL;
+	struct socket_info** list;
+	unsigned short c_proto;
+
+	c_proto = PROTO_UDP;
+	do {
+		/* get the proper sock_list */
+		list=get_sock_info_list(c_proto);
+
+		if (list==0) {
+			/* disabled or unknown protocol */
+			continue;
+		}
+
+		for (si=*list; si; si=si->next) {
+			if(si->useinfo.sock_str.s == NULL) {
+				continue;
+			}
+			LM_DBG("checking if listen %.*s matches %.*s\n",
+					sockstr->len, sockstr->s,
+					si->useinfo.sock_str.len, si->useinfo.sock_str.s);
+			if (sockstr->len == si->useinfo.sock_str.len
+					&& strncasecmp(sockstr->s, si->useinfo.sock_str.s,
+							sockstr->len)==0) {
+				return si;
+			}
+		}
+	} while((c_proto = next_proto(c_proto))!=0);
+
+	return NULL;
+}
+
+socket_info_t* ksr_get_socket_by_index(int idx)
+{
+	socket_info_t *si = NULL;
+	struct socket_info** list;
+	unsigned short c_proto;
+
+	if (idx<0) {
+		idx += _ksr_sockets_no;
+		if (idx < 0) {
+			LM_DBG("negative overall index\n");
+			return NULL;
+		}
+	}
+	c_proto = PROTO_UDP;
+	do {
+		/* get the proper sock_list */
+		list=get_sock_info_list(c_proto);
+
+		if (list==0) {
+			/* disabled or unknown protocol */
+			continue;
+		}
+
+		for (si=*list; si; si=si->next) {
+			if(idx==0) {
+				return si;
+			}
+			idx--;
+		}
+	} while((c_proto = next_proto(c_proto))!=0);
+
+	return NULL;
+}
+
 /* checks if the proto:port is one of the ports we listen on
  * and returns the corresponding socket_info structure.
  * if proto==0 (PROTO_NONE) the protocol is ignored
@@ -1070,6 +1290,7 @@ error:
 	req.g.rtgen_family = family;\
 	} while(0);
 
+#define NETLINK_BUFFER_SIZE 32768
 	
 static int get_flags(int family){
 	struct {
@@ -1079,12 +1300,12 @@ static int get_flags(int family){
 	int rtn = 0;
 	struct nlmsghdr*  nlp;
 	struct ifinfomsg *ifi;
-	char buf[8192];
+	char buf[NETLINK_BUFFER_SIZE];
 	char *p = buf;
 	int nll = 0;
     int nl_sock = -1;
 
-	fill_nl_req(req, RTM_GETLINK, AF_INET);
+	fill_nl_req(req, RTM_GETLINK, family);
 
 	if((nl_sock = nl_bound_sock()) < 0) return -1;
 
@@ -1095,6 +1316,10 @@ static int get_flags(int family){
 	}
 
 	while(1) {
+		if ((sizeof(buf) - nll) == 0) {
+			LM_ERR("netlink buffer overflow in get_flags");
+			goto error;
+		}
 		rtn = recv(nl_sock, p, sizeof(buf) - nll, 0);
 		nlp = (struct nlmsghdr *) p;
 		if(nlp->nlmsg_type == NLMSG_DONE){
@@ -1148,7 +1373,7 @@ static int build_iface_list(void)
 	struct nlmsghdr*  nlp;
 	struct ifaddrmsg *ifi;
 	int rtl;
-	char buf[8192];
+	char buf[NETLINK_BUFFER_SIZE];
 	char *p = buf;
 	int nll = 0;
 	struct rtattr * rtap;
@@ -1184,6 +1409,10 @@ static int build_iface_list(void)
 		nll = 0;
 		p = buf;
 		while(1) {
+			if ((sizeof(buf) - nll) == 0) {
+				LM_ERR("netlink buffer overflow in build_iface_list");
+				goto error;
+			}
 			rtn = recv(nl_sock, p, sizeof(buf) - nll, 0);
 			LM_DBG("received %d byles \n", rtn);
 			nlp = (struct nlmsghdr *) p;
@@ -1254,6 +1483,7 @@ static int build_iface_list(void)
 					case IFA_LABEL:
 						LM_DBG("iface name is %s\n", (char*)RTA_DATA(rtap));
 						strncpy(name, (char*)RTA_DATA(rtap), MAX_IF_LEN-1);
+						name[MAX_IF_LEN-1] = '\0';
 						break;
 					case IFA_BROADCAST:
 					case IFA_ANYCAST:
@@ -1272,7 +1502,8 @@ static int build_iface_list(void)
 			}
 
 			if(strlen(ifaces[index].name)==0 && strlen(name)>0) {
-				strncpy(ifaces[index].name, name, MAX_IF_LEN-1);
+				memcpy(ifaces[index].name, name, MAX_IF_LEN-1);
+				ifaces[index].name[MAX_IF_LEN-1] = '\0';
 			}
 
 			ifaces[index].index = index;
@@ -2062,10 +2293,11 @@ void print_all_socket_lists()
 				for (ai=si->addr_info_lst; ai; ai=ai->next) {
 					printf(", %s", ai->address_str.s);
 				}
-				printf("):%s%s%s\n",
+				printf("):%s%s%s%s\n",
 						si->port_no_str.s,
-						si->flags & SI_IS_MCAST ? " mcast" : "",
-						si->flags & SI_IS_MHOMED? " mhomed" : "");
+						si->flags & SI_IS_MCAST  ? " mcast" : "",
+						si->flags & SI_IS_MHOMED ? " mhomed" : "",
+						si->flags & SI_IS_VIRTUAL? " virtual" : "");
 			}else{
 				printf("             %s: %s",
 						get_valid_proto_name(proto),
@@ -2073,10 +2305,11 @@ void print_all_socket_lists()
 				if (!(si->flags & SI_IS_IP)) {
 					printf(" [%s]", si->address_str.s);
 				}
-				printf( ":%s%s%s",
+				printf( ":%s%s%s%s",
 						si->port_no_str.s,
-						si->flags & SI_IS_MCAST ? " mcast" : "",
-						si->flags & SI_IS_MHOMED? " mhomed" : "");
+						si->flags & SI_IS_MCAST  ? " mcast" : "",
+						si->flags & SI_IS_MHOMED ? " mhomed" : "",
+						si->flags & SI_IS_VIRTUAL? " virtual" : "");
 				if (si->sockname.s) {
 					printf(" name %s", si->sockname.s);
 				}
@@ -2189,11 +2422,14 @@ int parse_protohostport(str* ins, sr_phostp_t *r)
 	if (second) { /* 2 ':' found => check if valid */
 		if (parse_proto((unsigned char*)ins->s, first-ins->s, &r->proto)<0)
 			goto error_proto;
+		r->sproto.s = ins->s;
+		r->sproto.len = first-ins->s;
 
 		tmp.s=second+1;
 		tmp.len=(ins->s + ins->len) - tmp.s;
 
 		if (str2int(&tmp, (unsigned int *)&(r->port))<0) goto error_port;
+		r->sport = tmp;
 
 		r->host.s=first+1;
 		r->host.len=(int)(second-r->host.s);
@@ -2206,10 +2442,13 @@ int parse_protohostport(str* ins, sr_phostp_t *r)
 		/* invalid port => it's proto:host */
 		if (parse_proto((unsigned char*)ins->s, first-ins->s, &r->proto)<0)
 			goto error_proto;
+		r->sproto.s = ins->s;
+		r->sproto.len = first-ins->s;
 		r->host.s=first+1;
 		r->host.len=(int)(p-r->host.s);
 	}else{
 		/* valid port => its host:port */
+		r->sport = tmp;
 		r->host.s=ins->s;
 		r->host.len=(int)(first-r->host.s);
 	}
