@@ -83,6 +83,7 @@ int cb_dlg_cfg_reset(sip_msg_t *msg, unsigned int flags, void *cbp)
 	if(get_route_type()==LOCAL_ROUTE) {
 		return 1;
 	}
+
 	memset(&_dlg_ctx, 0, sizeof(dlg_ctx_t));
 
 	return 1;
@@ -115,7 +116,7 @@ static inline struct dlg_var *new_dlg_var(str *key, str *val)
 	var->key.len = key->len;
 	var->key.s = (char*)shm_malloc(var->key.len+1);
 	if (var->key.s==NULL) {
-		shm_free(var);			
+		shm_free(var);
 		LM_ERR("no more shm mem\n");
 		return NULL;
 	}
@@ -125,8 +126,8 @@ static inline struct dlg_var *new_dlg_var(str *key, str *val)
 	var->value.len = val->len;
 	var->value.s = (char*)shm_malloc(var->value.len+1);
 	if (var->value.s==NULL) {
-		shm_free(var->key.s);			
-		shm_free(var);			
+		shm_free(var->key.s);
+		shm_free(var);
 		LM_ERR("no more shm mem\n");
 		return NULL;
 	}
@@ -169,10 +170,10 @@ int set_dlg_variable_unsafe(struct dlg_cell *dlg, str *key, str *val)
 	struct dlg_var * it;
 	struct dlg_var * it_prev;
 	struct dlg_var ** var_list;
-	
-	if (dlg) 
+
+	if (dlg)
 		var_list = &dlg->vars;
-	else 
+	else
 		var_list = &_dlg_var_table;
 
 	if ( val && (var=new_dlg_var(key, val))==NULL) {
@@ -197,7 +198,7 @@ int set_dlg_variable_unsafe(struct dlg_cell *dlg, str *key, str *val)
 				/* Take the previous vflags: */
 				var->vflags = it->vflags | DLG_FLAG_CHANGED;
 				if (it_prev) it_prev->next = var;
-				else *var_list = var;				  
+				else *var_list = var;
 			}
 
 			/* Free this var: */
@@ -225,7 +226,7 @@ str * get_dlg_variable_unsafe(struct dlg_cell *dlg, str *key)
 {
 	struct dlg_var *var, *var_list;
 
-	if (dlg) 
+	if (dlg)
 		var_list = dlg->vars;
 	else
 		var_list = _dlg_var_table;
@@ -278,12 +279,15 @@ void print_lists(struct dlg_cell *dlg) {
 	}
 }
 
-str * get_dlg_variable(struct dlg_cell *dlg, str *key)
+/**
+ * return reference to the dlg variable value
+ * - unsafe - use only when it is sure that the value is not updated
+ */
+str* get_dlg_varref(struct dlg_cell *dlg, str *key)
 {
     str* var = NULL;
 
-    if( !dlg || !key || key->len > strlen(key->s))
-    {
+    if( !dlg || !key || key->len<=0) {
         LM_ERR("BUG - bad parameters\n");
 
         return NULL;
@@ -296,10 +300,139 @@ str * get_dlg_variable(struct dlg_cell *dlg, str *key)
     return var;
 }
 
+/**
+ * set *val to a pv buffer where the dlg value is cloned
+ * - safe to use immediately, before another dlg var get or other operation
+ *   done by the same process exceeds the number of pv buffer slots
+ */
+int get_dlg_varval(struct dlg_cell *dlg, str *key, str *val)
+{
+	str *var = NULL;
+
+	val->s = NULL;
+	val->len = 0;
+
+	if( !dlg || !key || key->len<=0) {
+		LM_ERR("BUG - bad parameters\n");
+		return -1;
+	}
+
+	dlg_lock(d_table, &(d_table->entries[dlg->h_entry]));
+	var = get_dlg_variable_unsafe(dlg, key);
+	if(var) {
+		val->len = pv_get_buffer_size();
+		if(val->len < var->len+1) {
+			LM_ERR("pv buffer too small (%d) - needed %d\n",
+					val->len, var->len+1);
+			val->s = NULL;
+			val->len = 0;
+			var = NULL;
+		} else {
+			val->s = pv_get_buffer();
+			memcpy(val->s, var->s, var->len);
+			val->len = var->len;
+			val->s[val->len] = '\0';
+		}
+	}
+	dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
+
+	if(var) {
+		return 0;
+	}
+	return -2;
+}
+
+/**
+ * set *val to a pkg-allocated buffer where the dlg value is cloned
+ * - val->s has to be pkg-freed after use
+ */
+int get_dlg_vardup(struct dlg_cell *dlg, str *key, str *val)
+{
+	str *var = NULL;
+
+	val->s = NULL;
+	val->len = 0;
+
+	if( !dlg || !key || key->len<=0) {
+		LM_ERR("BUG - bad parameters\n");
+		return -1;
+	}
+
+	dlg_lock(d_table, &(d_table->entries[dlg->h_entry]));
+	var = get_dlg_variable_unsafe(dlg, key);
+	if(var && var->s) {
+		val->s = (char*)pkg_malloc(var->len + 1);
+		if(val->s!=NULL) {
+			memcpy(val->s, var->s, var->len);
+			val->len = var->len;
+			val->s[val->len] = '\0';
+		}
+	}
+	dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
+
+	if(val->s) {
+		return 0;
+	}
+	return -2;
+}
+
+/**
+ * return the status if the dlg variable value is set or not
+ * - 1 - variable is set
+ * - 0 - variable is not set
+ */
+int get_dlg_varstatus(struct dlg_cell *dlg, str *key)
+{
+    str* var = NULL;
+    int ret = 0;
+
+    if( !dlg || !key || key->len<=0) {
+        LM_ERR("BUG - bad parameters\n");
+        return 0;
+    }
+
+    dlg_lock(d_table, &(d_table->entries[dlg->h_entry]));
+    var = get_dlg_variable_unsafe(dlg, key);
+	if(var && var->s) {
+		ret = 1;
+	}
+    dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
+
+    return ret;
+}
+
+int get_dlg_variable_uintval(struct dlg_cell *dlg, str *key, unsigned int *uval)
+{
+	str* var = NULL;
+
+	if( !dlg || !key || key->len <=0 || !uval) {
+		LM_ERR("BUG - bad parameters\n");
+		return -1;
+	}
+
+	dlg_lock(d_table, &(d_table->entries[dlg->h_entry]));
+	var = get_dlg_variable_unsafe(dlg, key);
+	if(var==NULL || var->s==NULL || var->len<=0) {
+		LM_DBG("no variable set yet\n");
+		goto error;
+	}
+	if(str2int(var, uval)<0) {
+		LM_ERR("invalid unsingned int value: %.*s\n",
+				var->len, var->s);
+		goto error;
+	}
+	dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
+	return 0;
+
+error:
+	dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
+	return -1;
+}
+
 int set_dlg_variable(struct dlg_cell *dlg, str *key, str *val)
 {
     int ret = -1;
-    if( !dlg || !key || key->len > strlen(key->s) || (val && val->len > strlen(val->s)))
+    if( !dlg || !key || !key->s || key->len<=0 )
     {
         LM_ERR("BUG - bad parameters\n");
         return -1;
@@ -325,6 +458,18 @@ done:
     return ret;
 }
 
+int set_dlg_variable_uintval(struct dlg_cell *dlg, str *key, unsigned int uval)
+{
+	str sval = STR_NULL;
+
+	sval.s = int2str(uval, &sval.len);
+	if(sval.s==NULL) {
+		return -1;
+	}
+
+	return set_dlg_variable(dlg, key, &sval);
+}
+
 int pv_get_dlg_variable(struct sip_msg *msg, pv_param_t *param, pv_value_t *res)
 {
 	dlg_cell_t *dlg;
@@ -394,7 +539,7 @@ int pv_set_dlg_variable(struct sip_msg* msg, pv_param_t *param, int op, pv_value
 
 	/* Retrieve the dialog for current message */
 	dlg=dlg_get_msg_dialog( msg);
-	
+
 	if (dlg) {
 		/* Lock the dialog */
 		dlg_lock(d_table, &(d_table->entries[dlg->h_entry]));
@@ -433,10 +578,11 @@ int pv_set_dlg_variable(struct sip_msg* msg, pv_param_t *param, int op, pv_value
 	/* unlock dialog */
 	if (dlg) {
 		dlg->dflags |= DLG_FLAG_CHANGED_VARS;
-		dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
-		if ( dlg_db_mode==DB_MODE_REALTIME )
+		if ( dlg_db_mode==DB_MODE_REALTIME ) {
+			/* dlg_lock() / dlg_unlock() are reentrant */
 			update_dialog_dbinfo(dlg);
-
+		}
+		dlg_unlock(d_table, &(d_table->entries[dlg->h_entry]));
 	}
 	print_lists(dlg);
 
@@ -464,8 +610,11 @@ int pv_get_dlg_ctx(struct sip_msg *msg,  pv_param_t *param,
 			return pv_get_uintval(msg, param, res,
 					(unsigned int)_dlg_ctx.to_bye);
 		case 4:
-			return pv_get_uintval(msg, param, res,
-					(unsigned int)_dlg_ctx.to_route);
+			if(_dlg_ctx.to_route>0) {
+				return pv_get_strzval(msg, param, res,
+						_dlg_ctx.to_route_name);
+			}
+			return pv_get_null(msg, param, res);
 		case 5:
 			_dlg_ctx.set = (_dlg_ctx.iuid.h_id==0)?0:1;
 			return pv_get_uintval(msg, param, res,
@@ -473,6 +622,9 @@ int pv_get_dlg_ctx(struct sip_msg *msg,  pv_param_t *param,
 		case 6:
 			return pv_get_uintval(msg, param, res,
 					(unsigned int)_dlg_ctx.dir);
+		case 7:
+			return pv_get_sintval(msg, param, res,
+					_dlg_ctx.to_route);
 		default:
 			return pv_get_uintval(msg, param, res,
 					(unsigned int)_dlg_ctx.on);
@@ -484,6 +636,7 @@ int pv_set_dlg_ctx(struct sip_msg* msg, pv_param_t *param,
 		int op, pv_value_t *val)
 {
 	int n;
+	int rlen;
 	char *rtp;
 
 	if(param==NULL)
@@ -510,13 +663,15 @@ int pv_set_dlg_ctx(struct sip_msg* msg, pv_param_t *param,
 						&& val->rs.len<DLG_TOROUTE_SIZE) {
 					_dlg_ctx.to_route = route_lookup(&main_rt, val->rs.s);
 					strcpy(_dlg_ctx.to_route_name, val->rs.s);
-				} else _dlg_ctx.to_route = 0;
+				} else { _dlg_ctx.to_route = 0; }
 			} else {
 				if(n!=0) {
-					rtp = int2str(n, NULL);
-					_dlg_ctx.to_route = route_lookup(&main_rt, rtp);
-					strcpy(_dlg_ctx.to_route_name, rtp);
-				} else _dlg_ctx.to_route = 0;
+					rtp = int2str(n, &rlen);
+					if(rlen<DLG_TOROUTE_SIZE) {
+						_dlg_ctx.to_route = route_lookup(&main_rt, rtp);
+						strcpy(_dlg_ctx.to_route_name, rtp);
+					} else { _dlg_ctx.to_route = 0; }
+				} else { _dlg_ctx.to_route = 0; }
 			}
 			if(_dlg_ctx.to_route <0) _dlg_ctx.to_route = 0;
 		break;
@@ -534,29 +689,29 @@ int pv_parse_dlg_ctx_name(pv_spec_p sp, str *in)
 
 	switch(in->len)
 	{
-		case 2: 
+		case 2:
 			if(strncmp(in->s, "on", 2)==0)
 				sp->pvp.pvn.u.isname.name.n = 0;
 			else goto error;
 		break;
-		case 3: 
+		case 3:
 			if(strncmp(in->s, "set", 3)==0)
 				sp->pvp.pvn.u.isname.name.n = 5;
 			else if(strncmp(in->s, "dir", 3)==0)
 				sp->pvp.pvn.u.isname.name.n = 6;
 			else goto error;
 		break;
-		case 5: 
+		case 5:
 			if(strncmp(in->s, "flags", 6)==0)
 				sp->pvp.pvn.u.isname.name.n = 1;
 			else goto error;
 		break;
-		case 7: 
+		case 7:
 			if(strncmp(in->s, "timeout", 7)==0)
 				sp->pvp.pvn.u.isname.name.n = 2;
 			else goto error;
 		break;
-		case 11: 
+		case 11:
 			if(strncmp(in->s, "timeout_bye", 11)==0)
 				sp->pvp.pvn.u.isname.name.n = 3;
 			else goto error;
@@ -566,6 +721,11 @@ int pv_parse_dlg_ctx_name(pv_spec_p sp, str *in)
 				sp->pvp.pvn.u.isname.name.n = 4;
 			else goto error;
 		break;
+		case 16:
+			if(strncmp(in->s, "timeout_route_id", 16)==0)
+				sp->pvp.pvn.u.isname.name.n = 7;
+			else goto error;
+		break;
 		default:
 			goto error;
 	}
@@ -586,6 +746,7 @@ int pv_get_dlg(struct sip_msg *msg, pv_param_t *param,
 	int res_type = 0;
 	str sv = { 0 };
 	unsigned int ui = 0;
+	int si = 0;
 
 	if(param==NULL)
 		return -1;
@@ -668,8 +829,8 @@ int pv_get_dlg(struct sip_msg *msg, pv_param_t *param,
 			sv.s[sv.len] = '\0';
 			break;
 		case 9:
-			res_type = 1;
-			ui = (unsigned int)dlg->toroute;
+			res_type = 3;
+			si = dlg->toroute;
 			break;
 		case 10:
 			if(dlg->cseq[DLG_CALLEE_LEG].s==NULL
@@ -802,6 +963,8 @@ done:
 			return pv_get_uintval(msg, param, res, ui);
 		case 2:
 			return pv_get_strval(msg, param, res, &sv);
+		case 3:
+			return pv_get_sintval(msg, param, res, si);
 		default:
 			return pv_get_null(msg, param, res);
 	}
@@ -814,24 +977,24 @@ int pv_parse_dlg_name(pv_spec_p sp, str *in)
 
 	switch(in->len)
 	{
-		case 3: 
+		case 3:
 			if(strncmp(in->s, "ref", 3)==0)
 				sp->pvp.pvn.u.isname.name.n = 0;
 			else goto error;
 		break;
-		case 4: 
+		case 4:
 			if(strncmp(in->s, "h_id", 4)==0)
 				sp->pvp.pvn.u.isname.name.n = 1;
 			else goto error;
 		break;
-		case 5: 
+		case 5:
 			if(strncmp(in->s, "state", 5)==0)
 				sp->pvp.pvn.u.isname.name.n = 2;
 			else if(strncmp(in->s, "to_rs", 5)==0)
 				sp->pvp.pvn.u.isname.name.n = 3;
 			else goto error;
 		break;
-		case 6: 
+		case 6:
 			if(strncmp(in->s, "dflags", 6)==0)
 				sp->pvp.pvn.u.isname.name.n = 4;
 			else if(strncmp(in->s, "sflags", 6)==0)
@@ -844,7 +1007,7 @@ int pv_parse_dlg_name(pv_spec_p sp, str *in)
 				sp->pvp.pvn.u.isname.name.n = 8;
 			else goto error;
 		break;
-		case 7: 
+		case 7:
 			if(strncmp(in->s, "toroute", 7)==0)
 				sp->pvp.pvn.u.isname.name.n = 9;
 			else if(strncmp(in->s, "to_cseq", 7)==0)
@@ -855,7 +1018,7 @@ int pv_parse_dlg_name(pv_spec_p sp, str *in)
 				sp->pvp.pvn.u.isname.name.n = 21;
 			else goto error;
 		break;
-		case 8: 
+		case 8:
 			if(strncmp(in->s, "from_uri", 8)==0)
 				sp->pvp.pvn.u.isname.name.n = 12;
 			else if(strncmp(in->s, "from_tag", 8)==0)
@@ -866,27 +1029,27 @@ int pv_parse_dlg_name(pv_spec_p sp, str *in)
 				sp->pvp.pvn.u.isname.name.n = 15;
 			else goto error;
 		break;
-		case 9: 
+		case 9:
 			if(strncmp(in->s, "from_cseq", 9)==0)
 				sp->pvp.pvn.u.isname.name.n = 16;
 			else goto error;
 		break;
-		case 10: 
+		case 10:
 			if(strncmp(in->s, "to_contact", 10)==0)
 				sp->pvp.pvn.u.isname.name.n = 17;
 			else goto error;
 		break;
-		case 11: 
+		case 11:
 			if(strncmp(in->s, "to_bindaddr", 11)==0)
 				sp->pvp.pvn.u.isname.name.n = 18;
 			else goto error;
 		break;
-		case 12: 
+		case 12:
 			if(strncmp(in->s, "from_contact", 12)==0)
 				sp->pvp.pvn.u.isname.name.n = 19;
 			else goto error;
 		break;
-		case 13: 
+		case 13:
 			if(strncmp(in->s, "from_bindaddr", 13)==0)
 				sp->pvp.pvn.u.isname.name.n = 20;
 			else goto error;
