@@ -34,12 +34,6 @@
  *  This module implement various functions and checks related to
  *  SIP message handling and URI handling.
  *
- *  It offers some functions related to handle ringing. In a
- *  parallel forking scenario you get several 183s with SDP. You
- *  don't want that your customers hear more than one ringtone or
- *  answer machine in parallel on the phone. So its necessary to
- *  drop the 183 in this cases and send a 180 instead.
- *
  *  This module provides a function to answer OPTIONS requests
  *  which are directed to the server itself. This means an OPTIONS
  *  request which has the address of the server in the request
@@ -65,8 +59,9 @@
 #include "../../core/error.h"
 #include "../../core/kemi.h"
 #include "../../core/parser/parse_option_tags.h"
+#include "../../core/parser/parse_uri.h"
+#include "../../core/parser/parse_date.h"
 
-#include "ring.h"
 #include "options.h"
 
 #include "checks.h"
@@ -94,8 +89,6 @@ str rpid_suffix = {DEF_RPID_SUFFIX, sizeof(DEF_RPID_SUFFIX) - 1};
 /*! Definition of AVP containing rpid value */
 char* rpid_avp_param = DEF_RPID_AVP;
 
-gen_lock_t *ring_lock = NULL;
-unsigned int ring_timeout = 0;
 /* for options functionality */
 str opt_accept = str_init(ACPT_DEF);
 str opt_accept_enc = str_init(ACPT_ENC_DEF);
@@ -107,6 +100,13 @@ sl_api_t opt_slb;
 static int mod_init(void);
 static void mod_destroy(void);
 
+static int w_contact_param_encode(sip_msg_t *msg, char *pnparam, char *psaddr);
+static int w_contact_param_decode(sip_msg_t *msg, char *pnparam, char *p2);
+static int w_contact_param_decode_ruri(sip_msg_t *msg, char *pnparam, char *p2);
+static int w_contact_param_rm(sip_msg_t *msg, char *pnparam, char *p2);
+
+static int w_hdr_date_check(sip_msg_t *msg, char *ptdiff, char *p2);
+
 /* Fixup functions to be defined later */
 static int fixup_set_uri(void** param, int param_no);
 static int fixup_free_set_uri(void** param, int param_no);
@@ -115,28 +115,33 @@ static int fixup_get_uri_param(void** param, int param_no);
 static int free_fixup_get_uri_param(void** param, int param_no);
 static int fixup_option(void** param, int param_no);
 
+static int ki_is_gruu(sip_msg_t *msg);
 
 char *contact_flds_separator = DEFAULT_SEPARATOR;
 
 static cmd_export_t cmds[]={
-	{"ring_insert_callid", (cmd_function)ring_insert_callid, 0, ring_fixup,
-		0, REQUEST_ROUTE|FAILURE_ROUTE},
 	{"options_reply",      (cmd_function)opt_reply,         0, 0,
 		0, REQUEST_ROUTE},
-	{"is_user",            (cmd_function)is_user,           1, fixup_str_null,
+	{"is_user",            (cmd_function)is_user,           1, fixup_spve_null,
 		0, REQUEST_ROUTE|LOCAL_ROUTE},
 	{"has_totag", 	       (cmd_function)w_has_totag,         0, 0,
 		0, ANY_ROUTE},
-	{"uri_param",          (cmd_function)uri_param_1,       1, fixup_str_null,
-		0, REQUEST_ROUTE|LOCAL_ROUTE},
-	{"uri_param",          (cmd_function)uri_param_2,       2, fixup_str_str,
-		0, REQUEST_ROUTE|LOCAL_ROUTE},
+	{"uri_param",          (cmd_function)uri_param_1,       1, fixup_spve_null,
+		0, REQUEST_ROUTE|BRANCH_ROUTE|FAILURE_ROUTE},
+	{"uri_param",          (cmd_function)uri_param_2,       2, fixup_spve_spve,
+		0, REQUEST_ROUTE|BRANCH_ROUTE|FAILURE_ROUTE},
+	{"uri_param_any",      (cmd_function)w_uri_param_any,   1, fixup_spve_null,
+		0, REQUEST_ROUTE|BRANCH_ROUTE|FAILURE_ROUTE},
 	{"add_uri_param",      (cmd_function)add_uri_param,     1, fixup_str_null,
 		0, REQUEST_ROUTE},
 	{"get_uri_param",      (cmd_function)get_uri_param,     2, fixup_get_uri_param,
 		free_fixup_get_uri_param, REQUEST_ROUTE|LOCAL_ROUTE},
+	{"uri_param_rm",       (cmd_function)w_uri_param_rm,    1, fixup_spve_null,
+		0, REQUEST_ROUTE|BRANCH_ROUTE|FAILURE_ROUTE},
 	{"tel2sip", (cmd_function)tel2sip, 3, fixup_tel2sip, 0,
 		REQUEST_ROUTE|FAILURE_ROUTE|BRANCH_ROUTE|ONREPLY_ROUTE},
+	{"is_uri",            (cmd_function)is_uri,           1, fixup_spve_null,
+		fixup_free_spve_null, ANY_ROUTE},
 	{"is_e164",            (cmd_function)is_e164,           1, fixup_pvar_null,
 		fixup_free_pvar_null, REQUEST_ROUTE|FAILURE_ROUTE|LOCAL_ROUTE},
 	{"is_uri_user_e164",   (cmd_function)w_is_uri_user_e164,  1, fixup_pvar_null,
@@ -151,6 +156,8 @@ static cmd_export_t cmds[]={
 		0, ANY_ROUTE},
 	{"cmp_aor",  (cmd_function)w_cmp_aor,                   2, fixup_spve_spve,
 		0, ANY_ROUTE},
+	{"cmp_hdr_name",  (cmd_function)w_cmp_hdr_name,         2, fixup_spve_spve,
+		0, ANY_ROUTE},
 	{"is_rpid_user_e164",   (cmd_function)is_rpid_user_e164, 0, 0,
 		0, REQUEST_ROUTE},
 	{"append_rpid_hf",      (cmd_function)append_rpid_hf,    0, 0,
@@ -161,8 +168,6 @@ static cmd_export_t cmds[]={
 		fixup_free_set_uri,	ANY_ROUTE},
 	{"set_uri_host", (cmd_function)set_uri_host,             2, fixup_set_uri,
 		fixup_free_set_uri,	ANY_ROUTE},
-	{"bind_siputils",       (cmd_function)bind_siputils,           1, 0,
-		0, 0},
 	{"is_request",          (cmd_function)w_is_request,            0, 0,
 		0, ANY_ROUTE},
 	{"is_reply",            (cmd_function)w_is_reply,              0, 0,
@@ -171,21 +176,40 @@ static cmd_export_t cmds[]={
 		0, ANY_ROUTE},
 	{"is_gruu",  (cmd_function)w_is_gruu,                    1, fixup_spve_null,
 		0, ANY_ROUTE},
-	{"is_supported",  (cmd_function)w_is_supported,                    1, fixup_option,
+	{"is_supported",  (cmd_function)w_is_supported,          1, fixup_option,
 		0, ANY_ROUTE},
-	{"is_first_hop",  (cmd_function)w_is_first_hop,                    0, 0,
+	{"is_first_hop",  (cmd_function)w_is_first_hop,          0, 0,
 		0, ANY_ROUTE},
+	{"is_first_hop",  (cmd_function)w_is_first_hop,          1, fixup_igp_null,
+		fixup_free_igp_null, ANY_ROUTE},
 	{"is_tel_number", (cmd_function)is_tel_number,           1, fixup_spve_null,
 		0, ANY_ROUTE},
 	{"is_numeric", (cmd_function)is_numeric,                 1, fixup_spve_null,
 		0, ANY_ROUTE},
+	{"is_alphanum", (cmd_function)ksr_is_alphanum,               1, fixup_spve_null,
+		0, ANY_ROUTE},
+	{"is_alphanumex", (cmd_function)ksr_is_alphanumex,           2, fixup_spve_spve,
+		0, ANY_ROUTE},
 	{"sip_p_charging_vector", (cmd_function)sip_handle_pcv,  1, fixup_spve_null,
 		fixup_free_spve_null, ANY_ROUTE},
+	{"contact_param_encode",      (cmd_function)w_contact_param_encode,    2,
+		fixup_spve_spve, fixup_free_spve_spve, REQUEST_ROUTE|ONREPLY_ROUTE},
+	{"contact_param_decode",      (cmd_function)w_contact_param_decode,    1,
+		fixup_spve_null, fixup_free_spve_null, REQUEST_ROUTE|ONREPLY_ROUTE},
+	{"contact_param_decode_ruri", (cmd_function)w_contact_param_decode_ruri, 1,
+		fixup_spve_null, fixup_free_spve_null, REQUEST_ROUTE},
+	{"contact_param_rm",      (cmd_function)w_contact_param_rm,    1,
+		fixup_spve_null, fixup_free_spve_null, REQUEST_ROUTE|ONREPLY_ROUTE},
+	{"hdr_date_check",  (cmd_function)w_hdr_date_check,      1, fixup_igp_null,
+		fixup_free_igp_null, ANY_ROUTE},
+
+	{"bind_siputils",       (cmd_function)bind_siputils,           1, 0,
+		0, 0},
+
 	{0,0,0,0,0,0}
 };
 
 static param_export_t params[] = {
-	{"ring_timeout",            INT_PARAM, &default_siputils_cfg.ring_timeout},
 	{"options_accept",          PARAM_STR, &opt_accept},
 	{"options_accept_encoding", PARAM_STR, &opt_accept_enc},
 	{"options_accept_language", PARAM_STR, &opt_accept_lang},
@@ -206,38 +230,21 @@ static pv_export_t mod_pvs[] =  {
 };
 
 struct module_exports exports= {
-	"siputils",
+	"siputils",      /* module name */
 	DEFAULT_DLFLAGS, /* dlopen flags */
-	cmds,            /* Exported functions */
+	cmds,            /* exported functions */
 	params,          /* param exports */
-	0,               /* exported statistics */
-	0,               /* exported MI functions */
+	0,               /* exported RPC functions */
 	mod_pvs,         /* exported pseudo-variables */
-	0,               /* extra processes */
+	0,               /* response function */
 	mod_init,        /* initialization function */
-	0,               /* Response function */
-	mod_destroy,     /* Destroy function */
-	0,               /* Child init function */
+	0,               /* child init function */
+	mod_destroy      /* destroy function */
 };
 
 
 static int mod_init(void)
 {
-	if(default_siputils_cfg.ring_timeout > 0) {
-		ring_init_hashtable();
-
-		ring_lock = lock_alloc();
-		assert(ring_lock);
-		if (lock_init(ring_lock) == 0) {
-			LM_CRIT("cannot initialize lock.\n");
-			return -1;
-		}
-		if (register_script_cb(ring_filter, PRE_SCRIPT_CB|ONREPLY_CB, 0) != 0) {
-			LM_ERR("could not insert callback");
-			return -1;
-		}
-	}
-
 	/* bind the SL API */
 	if (sl_load_api(&opt_slb)!=0) {
 		LM_ERR("cannot bind to SL API\n");
@@ -260,13 +267,7 @@ static int mod_init(void)
 
 static void mod_destroy(void)
 {
-	if (ring_lock) {
-		lock_destroy(ring_lock);
-		lock_dealloc((void *)ring_lock);
-		ring_lock = NULL;
-	}
 
-	ring_destroy_hashtable();
 }
 
 
@@ -453,9 +454,139 @@ static int fixup_option(void** param, int param_no) {
 	return 0;
 }
 
+static int w_contact_param_encode(sip_msg_t *msg, char *pnparam, char *psaddr)
+{
+	str nparam = STR_NULL;
+	str saddr = STR_NULL;
+
+	if(fixup_get_svalue(msg, (gparam_t*)pnparam, &nparam)<0) {
+		LM_ERR("failed to get p1\n");
+		return -1;
+	}
+
+	if(fixup_get_svalue(msg, (gparam_t*)psaddr, &saddr)<0) {
+		LM_ERR("failed to get p1\n");
+		return -1;
+	}
+	return ki_contact_param_encode(msg, &nparam, &saddr);
+}
+
+static int w_contact_param_decode(sip_msg_t *msg, char *pnparam, char *p2)
+{
+	str nparam = STR_NULL;
+
+	if(fixup_get_svalue(msg, (gparam_t*)pnparam, &nparam)<0) {
+		LM_ERR("failed to get p1\n");
+		return -1;
+	}
+
+	return ki_contact_param_decode(msg, &nparam);
+}
+
+static int w_contact_param_decode_ruri(sip_msg_t *msg, char *pnparam, char *p3)
+{
+	str nparam = STR_NULL;
+
+	if(fixup_get_svalue(msg, (gparam_t*)pnparam, &nparam)<0) {
+		LM_ERR("failed to get p1\n");
+		return -1;
+	}
+
+	return ki_contact_param_decode_ruri(msg, &nparam);
+}
+
+static int w_contact_param_rm(sip_msg_t *msg, char *pnparam, char *p2)
+{
+	str nparam = STR_NULL;
+
+	if(fixup_get_svalue(msg, (gparam_t*)pnparam, &nparam)<0) {
+		LM_ERR("failed to get p1\n");
+		return -1;
+	}
+
+	return ki_contact_param_rm(msg, &nparam);
+}
+
+/*
+ * Check if pseudo variable contains a valid uri
+ */
+static int ki_is_uri(sip_msg_t* msg, str* suri)
+{
+	sip_uri_t turi;
+
+	if(suri==NULL || suri->s==NULL || suri->len<=0) {
+		return -1;
+	}
+	if(parse_uri(suri->s, suri->len, &turi)!=0) {
+		return -1;
+	}
+	return 1;
+}
+
+/*
+ * Check date header value with time difference
+ */
+static int ki_hdr_date_check(sip_msg_t* msg, int tdiff)
+{
+	time_t tnow, tmsg;
+
+	if ((!msg->date) && (parse_headers(msg, HDR_DATE_F, 0) == -1)) {
+		LM_ERR("failed parsing Date header\n");
+		return -1;
+	}
+	if (!msg->date) {
+		LM_ERR("Date header field is not found\n");
+		return -1;
+	}
+	if ((!(msg->date)->parsed) && (parse_date_header(msg) < 0)) {
+		LM_ERR("failed parsing DATE body\n");
+		return -1;
+	}
+
+#ifdef HAVE_TIMEGM
+	tmsg=timegm(&get_date(msg)->date);
+#else
+	tmsg=_timegm(&get_date(msg)->date);
+#endif
+	if (tmsg < 0) {
+		LM_ERR("timegm error\n");
+		return -2;
+	}
+
+	if ((tnow=time(0)) < 0) {
+		LM_ERR("time error %s\n", strerror(errno));
+		return -3;
+	}
+
+	if (tnow > tmsg + tdiff) {
+		LM_ERR("autdated date header value (%ld sec)\n", tnow - tmsg + tdiff);
+		return -4;
+	} else {
+		LM_ERR("Date header value OK\n");
+	}
+
+	return 1;
+
+}
+
 /**
  *
  */
+static int w_hdr_date_check(sip_msg_t *msg, char *ptdiff, char *p2)
+{
+	int tdiff = 0;
+
+	if(fixup_get_ivalue(msg, (gparam_t*)ptdiff, &tdiff)<0) {
+		LM_ERR("failed to get time diff parameter\n");
+		return -1;
+	}
+	return ki_hdr_date_check(msg, tdiff);
+}
+
+/**
+ *
+ */
+/* clang-format off */
 static sr_kemi_t sr_kemi_siputils_exports[] = {
 	{ str_init("siputils"), str_init("has_totag"),
 		SR_KEMIP_INT, has_totag,
@@ -477,9 +608,134 @@ static sr_kemi_t sr_kemi_siputils_exports[] = {
 		{ SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE,
 			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
 	},
+	{ str_init("siputils"), str_init("is_first_hop_mode"),
+		SR_KEMIP_INT, is_first_hop_mode,
+		{ SR_KEMIP_INT, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_uri"),
+		SR_KEMIP_INT, ki_is_uri,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_user"),
+		SR_KEMIP_INT, ki_is_user,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("uri_param"),
+		SR_KEMIP_INT, ki_uri_param,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("uri_param_value"),
+		SR_KEMIP_INT, ki_uri_param_value,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("uri_param_any"),
+		SR_KEMIP_INT, ki_uri_param_any,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("uri_param_rm"),
+		SR_KEMIP_INT, ki_uri_param_rm,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_tel_number"),
+		SR_KEMIP_INT, ki_is_tel_number,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_numeric"),
+		SR_KEMIP_INT, ki_is_numeric,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_alphanum"),
+		SR_KEMIP_INT, ki_is_alphanum,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_alphanumex"),
+		SR_KEMIP_INT, ki_is_alphanumex,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("options_reply"),
+		SR_KEMIP_INT, ki_opt_reply,
+		{ SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("encode_contact"),
+		SR_KEMIP_INT, ki_encode_contact,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("decode_contact"),
+		SR_KEMIP_INT, ki_decode_contact,
+		{ SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("decode_contact_header"),
+		SR_KEMIP_INT, ki_decode_contact_header,
+		{ SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("contact_param_encode"),
+		SR_KEMIP_INT, ki_contact_param_encode,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("contact_param_decode"),
+		SR_KEMIP_INT, ki_contact_param_decode,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("contact_param_decode_ruri"),
+		SR_KEMIP_INT, ki_contact_param_decode_ruri,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("contact_param_rm"),
+		SR_KEMIP_INT, ki_contact_param_rm,
+		{ SR_KEMIP_STR, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("hdr_date_check"),
+		SR_KEMIP_INT, ki_hdr_date_check,
+		{ SR_KEMIP_INT, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("cmp_uri"),
+		SR_KEMIP_INT, ki_cmp_uri,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("cmp_aor"),
+		SR_KEMIP_INT, ki_cmp_aor,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("cmp_hdr_name"),
+		SR_KEMIP_INT, ki_cmp_hdr_name,
+		{ SR_KEMIP_STR, SR_KEMIP_STR, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
+	{ str_init("siputils"), str_init("is_gruu"),
+		SR_KEMIP_INT, ki_is_gruu,
+		{ SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE,
+			SR_KEMIP_NONE, SR_KEMIP_NONE, SR_KEMIP_NONE }
+	},
 
 	{ {0, 0}, {0, 0}, 0, NULL, { 0, 0, 0, 0, 0, 0 } }
 };
+/* clang-format on */
+
+static int ki_is_gruu(sip_msg_t *msg) {
+	return w_is_gruu(msg, NULL, NULL);
+}
 
 /**
  *
