@@ -160,9 +160,9 @@ static inline unsigned int dlg2hash( dlg_t* dlg )
 }
 
 /**
- * refresh hdr shortcuts inside new buffer
+ * refresh r-uri and hdr shortcuts inside new buffer
  */
-int uac_refresh_hdr_shortcuts(tm_cell_t *tcell, char *buf, int buf_len)
+int uac_refresh_shortcuts(tm_cell_t *tcell, int branch, char *buf, int buf_len)
 {
 	sip_msg_t lreq;
 	struct cseq_body *cs;
@@ -175,6 +175,7 @@ int uac_refresh_hdr_shortcuts(tm_cell_t *tcell, char *buf, int buf_len)
 		LM_ERR("failed to parse headers in new message\n");
 		goto error;
 	}
+	tcell->uac[branch].uri = *GET_RURI(&lreq);
 	tcell->from.s = lreq.from->name.s;
 	tcell->from.len = lreq.from->len;
 	tcell->to.s = lreq.to->name.s;
@@ -300,10 +301,6 @@ static inline int t_run_local_req(
 	set_route_type( backup_route_type );
 	p_onsend=0;
 
-	/* restore original environment */
-	tm_xdata_swap(new_cell, &backup_xd, 1);
-	setsflagsval(sflag_bk);
-
 	if (unlikely(ra_ctx.run_flags&DROP_R_F)) {
 		LM_DBG("tm:local-request dropped msg. to %.*s\n",
 				lreq.dst_uri.len, lreq.dst_uri.s);
@@ -372,6 +369,10 @@ normal_update:
 	}
 
 clean:
+	/* restore original environment */
+	tm_xdata_swap(new_cell, &backup_xd, 1);
+	setsflagsval(sflag_bk);
+
 	/* clean local msg structure */
 	if (unlikely(lreq.new_uri.s))
 	{
@@ -541,13 +542,13 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 		if (unlikely(refresh_shortcuts==E_DROP)) {
 			ret=E_DROP;
 			goto error1;
-		}			
+		}
 	}
 #endif
 
 #ifdef USE_DNS_FAILOVER
 	/* Set the outgoing message as UAS, so the failover code has something to work with */
-	if(cfg_get(core, core_cfg, use_dns_failover)) {
+	if(cfg_get(core, core_cfg, use_dns_failover) || cfg_get(tm, tm_cfg, enable_uac_fr)) {
 		if(likely(t_build_msg_from_buf(&lreq, buf, buf_len, uac_r, &dst) == 0)) {
 			if (parse_headers(&lreq, HDR_EOH_F, 0) == -1) {
 				LM_ERR("failed to parse headers on uas for failover\n");
@@ -576,7 +577,7 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 	request->buffer = buf;
 	request->buffer_len = buf_len;
 	if(unlikely(refresh_shortcuts==1)) {
-		if(uac_refresh_hdr_shortcuts(new_cell, buf, buf_len)<0) {
+		if(uac_refresh_shortcuts(new_cell, 0, buf, buf_len)<0) {
 			LM_ERR("failed to refresh header shortcuts\n");
 			goto error1;
 		}
